import java.awt.Color;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Random;

import jas.engine.Sim;
import jas.events.ISimEventListener;
import jas.graph.RelationalAgent;

/**
 *
 * <p>Title: JAS</p>
 * <p>Description: Java Agent-based Simulation library</p>
 * <p>Copyright (C) 2002 Michele Sonnessa</p>
 *
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * @author Michele Sonnessa
 *
 */

public class Device extends RelationalAgent implements ISimEventListener{
	public final static int PROTOCOL_ACTIVITY = 0;
	public final static int COMMUNICATE = 1;
	public final static int CHECK_CONNECTION = 2;

	public final static int STATE_MASTER = 0;
	public final static int STATE_SLAVE = 1;
	public final static int STATE_INQUIRY = 3;
	public final static int STATE_BRIDGE = 2;
	public final static int STATE_BROKEN = 4;
	private int state;

	public final static int FREE_SLOTS = 7;

	public ArrayList offers, offerRequests;
	public FileTransfer fileTransfer;

	public Device master;
	public ArrayList gateways, piconetMembers;
	public int freeSlots = FREE_SLOTS;

	public int waitBeforeInquiry;
	public boolean enablePrice;

	private boolean scheduleGateAuction = false, scheduleAuction = false;
	private Random rnd;

	private DeviceMap deviceMap;

	/**
	 * @param arg0
	 * @param arg1
	 */
	public Device(DeviceMap map, int arg1, int waitBeforeInquiry, boolean enablePrice) {
		super(map, arg1);
		offers = new ArrayList();
		offerRequests = new ArrayList();
		master = null;
		gateways = new ArrayList();
		piconetMembers = new ArrayList();
		setState(STATE_BROKEN);
		setSize(12);
		this.waitBeforeInquiry = waitBeforeInquiry;
		this.enablePrice = enablePrice;

		rnd = new Random(Sim.getRnd().getLongFromTo(0, 123456));
	}

	public void postMeAnOffer(Device device)
	{
		offerRequests.add(device);
	}

	private DeviceInfo getBestMasterOffer()
	{
		DeviceInfo best = null;

		java.util.Collections.shuffle(offers, rnd);
		for (Iterator iter = offers.iterator(); iter.hasNext();) {
			DeviceInfo devInfo = (DeviceInfo) iter.next();
			if (devInfo.price < 100 && devInfo.device.getState() == STATE_MASTER)
			{
				if (best == null)
					best = devInfo;
				else if (best.price > devInfo.price)
					best = devInfo;
			}
		}

		return best;
	}

	private DeviceInfo getBestGateOffer()
	{
		DeviceInfo best = null;

		java.util.Collections.shuffle(offers, rnd);
		for (Iterator iter = offers.iterator(); iter.hasNext();) {
			DeviceInfo devInfo = (DeviceInfo) iter.next();
			if (devInfo.price < 100 && devInfo.device.getState() == STATE_SLAVE &&
					devInfo.device.getMaster() != this)
			{
				if (best == null)
					best = devInfo;
				else if (best.price > devInfo.price)
					best = devInfo;
			}
		}

		return best;
	}

	public void decideRole()
	{
		if (state != STATE_INQUIRY)
			return;

		DeviceInfo best = getBestMasterOffer();
		if (best == null)
			becomeMaster();
		else
			becomeSlave(best);
	}

	public boolean acceptSlave(Device device)
	{
		if (freeSlots <= 0 || state != STATE_MASTER)
			return false;

		piconetMembers.add(device);
		freeSlots--;
		this.addRelation(device, 0.5);

		return true;
	}

	private void setState(int newState)
	{
		switch (newState)
		{
			case STATE_BROKEN:
				setColor(Color.gray);
				setShape(RelationalAgent.SHAPE_CIRCLE);
				offers.clear();
				offerRequests.clear();
				for (Iterator iter = gateways.iterator(); iter.hasNext();) {
					Device dev = (Device) iter.next();
					removeRelation(dev);
					dev.removeRelation(this);
					iter.remove();
				}
				for (Iterator iter = piconetMembers.iterator(); iter.hasNext();) {
					Device dev = (Device) iter.next();
					removeRelation(dev);
					dev.removeRelation(this);
					iter.remove();
				}
				freeSlots = FREE_SLOTS;
				break;
			case STATE_INQUIRY:
				setColor(Color.yellow);
				setShape(RelationalAgent.SHAPE_CIRCLE);
				offers.clear();
				offerRequests.clear();
				for (Iterator iter = gateways.iterator(); iter.hasNext();) {
					Device dev = (Device) iter.next();
					removeRelation(dev);
					dev.removeRelation(this);
					iter.remove();
				}
				for (Iterator iter = piconetMembers.iterator(); iter.hasNext();) {
					Device dev = (Device) iter.next();
					removeRelation(dev);
					dev.removeRelation(this);
					iter.remove();
				}
				freeSlots = FREE_SLOTS;
				break;
			case STATE_MASTER:
			if (Sim.getAbsoluteTime() > 1941 && getLabel().equals("N1"))
										System.out.println("");

				setColor(Color.red);
				setShape(RelationalAgent.SHAPE_RECTANGLE);
				break;
			case STATE_SLAVE:
				setColor(Color.blue);
				piconetMembers.clear();
				setShape(RelationalAgent.SHAPE_CIRCLE);
				break;
			case STATE_BRIDGE:
				setColor(Color.green);
				piconetMembers.clear();
				setShape(RelationalAgent.SHAPE_TRIANGLE);
		}
		state = newState;
	}

	private void becomeMaster()
	{
		setState(STATE_MASTER);
	}

	public void becomeBridge(Device mast)
	{
		gateways.add(mast);
		master.gateways.add(this);
		freeSlots--;
		setState(STATE_BRIDGE);
		addRelation(mast, 0.3);
	}

	private void becomeSlave(DeviceInfo best)
	{
		if (graph.containsVertex(best.device))
			if (best.device.acceptSlave(this))
			{
				master = best.device;
				gateways.add(master);
				this.addRelation(master, 0.5);
				setState(STATE_SLAVE);
				return;
			}
		becomeMaster();
	}

	private void auction()
	{
		ArrayList list = ((DeviceMap) graph).getReachableDevicesFrom(this);
		for (Iterator iter = list.iterator(); iter.hasNext();) {
			Device device = (Device) iter.next();
			device.postMeAnOffer(this);
		}
		Sim.engine.getEventList().scheduleSimple(Sim.getAbsoluteTime() + 10, this, "decideRole");

		setState(STATE_INQUIRY);
	}

	public void protocolActivity()
	{
		if (scheduleAuction)
		{
			scheduleAuction=false;
			auction();
			return;
		}

		switch (state)
		{
			case STATE_INQUIRY:
				return;
			case STATE_BROKEN:
				auction();
				break;
			case STATE_MASTER:
				if (gateways.size() < ((DeviceMap) graph).numberOfGatewayPerMaster || scheduleGateAuction)
					searchForBridge();
				scheduleGateAuction = false;
		}

		for (Iterator iter = offerRequests.iterator(); iter.hasNext();) {
			Device dev = (Device) iter.next();
			if (graph.containsVertex(dev))
				dev.receiveAnOffer(getOffer());
			iter.remove();
		}
	}

	public void searchForBridge()
	{
		if (!graph.containsVertex(this))
			return;


		DeviceInfo best = getBestGateOffer();
		if (best == null)
			return;

		if (graph.containsVertex(best.device) && best.device.getMaster() != this)
		{
			gateways.add(best.device);
			freeSlots--;
			addRelation(best.device, 0.3);
			best.device.becomeBridge(this);
		}
	}

	public void checkConnections()
	{
		switch (state)
		{
			case STATE_SLAVE:
				if (!graph.containsVertex(master) || master.getState() != STATE_MASTER)
					setState(STATE_BROKEN);
				break;

			case STATE_BRIDGE:
				if (!graph.containsVertex(master) || master.getState() != STATE_MASTER)
				{
					setState(STATE_BROKEN);
					return;
				}

				for (Iterator iter = gateways.iterator(); iter.hasNext();) {
					Device dev = (Device) iter.next();
					if (!graph.containsVertex(dev) || dev.getState() != STATE_MASTER)
					{
						iter.remove();
						freeSlots++;
					}
				}
				if (gateways.size() == 1)
					setState(STATE_SLAVE);
				break;

			case STATE_MASTER:
				for (Iterator iter = piconetMembers.iterator(); iter.hasNext();) {
					Device element = (Device) iter.next();
					if (element != null && !graph.containsVertex(element))
					{
						iter.remove();
						freeSlots++;
					}
				}
				for (Iterator iter = gateways.iterator(); iter.hasNext();) {
					Device dev = (Device) iter.next();
					if (!graph.containsVertex(dev)  || dev.getState() != STATE_BRIDGE)
					{
						iter.remove();
						freeSlots++;
					}
				}
		}
	}

	public void receiveAnOffer(DeviceInfo offer)
	{
		offers.add(offer);
	}

	private DeviceInfo getOffer()
	{
		DeviceInfo info = new DeviceInfo();
		info.device = this;
		if (enablePrice)
		{

			info.price = 100.0 * (0.2 * Math.exp(getSlotSaturationComponent())) +
									 100.0 * (0.4 * Math.exp(getTrafficLoadComponent()));

		}
//			info.price = 100.0 * (Math.exp(getSlotSaturationComponent()) - 1.0) ;//+
//									 //100.0 * (Math.exp(0.6 * getTrafficLoadComponent()) - 1.0);
		else
			info.price = 80;

if (info.price != 80.0)
System.out.println("p " + info.price + " ");
		return info;
	}

	public void communicate()
	{

		if (fileTransfer == null)
			return;

		fileTransfer.beginRoute();
		boolean res;
		if (state != STATE_MASTER)
		{
			if (master == null)
				res = false;
			else
				res = master.routeForward(fileTransfer);
		}
		else
			res = routeForward(fileTransfer);

		if (res)
			fileTransfer.routeCompleted();
		else
			fileTransfer.routeFailed();

//		if (fileTransfer.getFailed() > 0 && fileTransfer.getFailed() % 10 == 0)
//			System.out.println("Fail " + this.getLabel() + " to " + fileTransfer.getTarget().getLabel());

		if (fileTransfer.getFailed() > waitBeforeInquiry)
		{
			fileTransfer.reconnections++;
			fileTransfer.routeCompleted();
			scheduleAuction = true;
			return;
		}

		if (fileTransfer.getPacketsToSend() <= 0 || fileTransfer.reconnections > 5)
			fileTransfer = null;

	}

	public boolean routeForward(FileTransfer file)
	{
		if (file.hasVisited(this))
			return false;

		file.hopOnNode(this);
		if (file.getTarget() == this || piconetMembers.contains(file.getTarget()))
			return true;

		for (Iterator iter = gateways.iterator(); iter.hasNext();) {
			Device gate = (Device) iter.next();
			if (gate.routeForward(file))
				return true;
		}

		if (file.getPath().size() < 4)
			scheduleGateAuction = true;

		for (Iterator iter = offers.iterator(); iter.hasNext();) {
			DeviceInfo info = (DeviceInfo) iter.next();
			if (info.device == file.getTarget() && state == STATE_MASTER )
				if (graph.containsVertex(info.device) && info.device.getMaster() != this)
				{
					scheduleGateAuction = true;
				}
			}

		if (!scheduleGateAuction)
			for (Iterator iter = offers.iterator(); iter.hasNext();) {
				DeviceInfo info = (DeviceInfo) iter.next();
				if (state == STATE_SLAVE )
					if (graph.containsVertex(info.device) && info.device.doYouKnow(file.getTarget()))
						info.device.auction();
			}

		return false;
	}

	public boolean doYouKnow(Device device)
	{
		for (Iterator iter = offers.iterator(); iter.hasNext();) {
			DeviceInfo info = (DeviceInfo) iter.next();
			if (info.device == device)
				if (graph.containsVertex(info.device) )
					return true;
			}
		return false;
	}

	public void setFileTransfer(FileTransfer file)
	{
		fileTransfer = file;
	}

	public int hashCode()
	{
		return id;
	}

	/* (non-Javadoc)
	 * @see jas.events.ISimEventListener#performAction(int)
	 */
	public void performAction(int actionId) {
		switch (actionId)
		{
			case PROTOCOL_ACTIVITY:
				protocolActivity();
				break;
			case COMMUNICATE:
				communicate();
				break;
			case CHECK_CONNECTION:
				checkConnections();
				break;
			default:
				throw new UnsupportedOperationException("Device agent supports only PROTOCOL_ACTIVITY and COMMUNICATE events.");
		}
	}

	private double getTrafficLoadComponent()
	{
		double sum = 0;
		for (Iterator iter = piconetMembers.iterator(); iter.hasNext();) {
			Device dev = (Device) iter.next();
			if (dev.getFileTransfer() != null)
				sum += dev.getFileTransfer().getPacketsToSend();
		}

		return sum / 7000.0;
	}

	private double getSlotSaturationComponent()
	{
		if (freeSlots < 0)
			freeSlots = 0;

		return (double)(FREE_SLOTS - freeSlots) / (double)FREE_SLOTS;
	}

	/**
	 * @return
	 */
	public int getState() {
		return state;
	}

	/**
	 * @return
	 */
	public Device getMaster() {
		return master;
	}

	public class DeviceInfo
	{
		Device device;
		double price;
	}
	/**
	 * @return
	 */
	public FileTransfer getFileTransfer() {
		return fileTransfer;
	}


	public DeviceMap getDeviceMap() {
		return deviceMap;
	}

	public void setDeviceMap(DeviceMap deviceMap) {
		this.deviceMap = deviceMap;
	}

}
