import java.util.ArrayList;
import java.util.Iterator;

import jas.engine.Sim;

import org._3pq.jgrapht.EdgeFactory;
import org._3pq.jgrapht.graph.DirectedWeightedMultigraph;


/**
 *
 * <p>Title: JAS</p>
 * <p>Description: Java Agent-based Simulation library</p>
 * <p>Copyright (C) 2002 Michele Sonnessa</p>
 *
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation;
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * @author Michele Sonnessa
 *
 */
public class DeviceMap extends DirectedWeightedMultigraph {
	public final float DISTANCE = 150;
	public int numberOfGatewayPerMaster;
	private int xSize, ySize;

	private Device device;

	/**
	 *
	 */
	public DeviceMap(int xSize, int ySize, int numberOfGatewayPerMaster) {
		super();

		this.xSize = xSize;
		this.ySize = ySize;
		this.numberOfGatewayPerMaster = numberOfGatewayPerMaster;
	}

	/**
	 * @param arg0
	 */
	public DeviceMap(EdgeFactory arg0) {
		super(arg0);
	}

	public void placeDeviceRandomly(Device device)
	{
		int x = Sim.getRnd().getIntFromTo(0, xSize);
		int y = Sim.getRnd().getIntFromTo(0, ySize);
		device.setXY(x, y);
	}
	/**
	 * @return
	 */
	public int getXSize() {
		return xSize;
	}

	/**
	 * @return
	 */
	public int getYSize() {
		return ySize;
	}

	public ArrayList getReachableDevicesFrom(Device device)
	{
	  Device candidate;
	  double distance;
	  ArrayList list = new ArrayList();

	  Iterator it = vertexSet().iterator();

	  while (it.hasNext())
	  {
			candidate = (Device) it.next();
			int x = device.getX() - candidate.getX();
			int y = device.getY() - candidate.getY();
			distance = Math.sqrt(x*x + y*y);

			if (distance < DISTANCE && device != candidate)
			  list.add(candidate);
	  }

	  return list;
	}


	public Device getDevice() {
		return device;
	}

	public void setDevice(Device device) {
		this.device = device;
	}

}
